@@ -76,10 +76,11 @@ def h_theorem(name):
 def h_newtheorem(parser, buf, mac, args, pos):
     name = parser.get_text_expanded(args[0])
     title = parser.get_text_expanded(args[2])
-    def f(parms):
+    def f(parser, options):
+        parms = parser.parms
         envs = [defs.Environ(parms, name, args='O', repl=h_theorem(title))]
-        return defs.ModParm(environments=envs)
-    parser.modify_parameters(f)
+        return defs.InitModule(environments=envs)
+    parser.modify_parameters(f, [])
     return []
 
 #   heading macros: append '.', unless last char in parms.heading_punct
@@ -114,7 +115,7 @@ def h_cite(parser, buf, mac, args, pos):
     return out
 
 #   macro \LTinclude: read macro definitions from file
-#   - this also activates packages
+#   - this also activates packages and switches languages
 #
 def h_load_defs(parser, buf, mac, args, pos):
     if not parser.read_macros:
@@ -124,16 +125,22 @@ def h_load_defs(parser, buf, mac, args, pos):
     if not ok:
         return utils.latex_error('could not read file ' + repr(file),
                                         pos, parser.latex, parser.parms)
-    parser.parser_work(latex)
-    return []
+    toks = parser.parser_work(latex)
+    return utils.filter_set_toks(toks, pos, defs.LanguageToken)
 
 #   read definitions for a LaTeX package
 #
 def h_load_module(prefix):
     def f(parser, buf, mac, args, pos):
-        pack = parser.get_text_expanded(args[1])
-        f = utils.get_module_handler(pack, prefix)
-        parser.init_package(pack, f)
-        return []
+        options = parser.parse_keyvals_list(args[0])
+        options = parser.expand_keyvals(options)
+        packs = parser.get_text_expanded(args[1])
+        out = []
+        for p in packs.split(','):
+            p = p.strip()
+            if p:
+                f = utils.get_module_handler(p, prefix)
+                out += parser.init_package(p, f, options)
+        return utils.filter_set_toks(out, pos, None)
     return f
 
