@@ -28,7 +28,7 @@ class Parser:
     def __init__(self, parms, packages=[], read_macros=None):
         self.parms = parms
         self.read_macros = read_macros
-        self.packages = []
+        self.packages = {}
         self.the_macros = {}
         self.the_environments = {}
         self.mathparser = mathparser.MathParser(self)
@@ -44,75 +44,101 @@ class Parser:
                 yield parms.item_default_label[0]
         self.item_lab_stack = [(labs_default(0), '')]
 
-        # for expansion of verbatim environment
-        self.verbatim_begin = [
-                    defs.BeginToken(0, '\\begin'),
-                    defs.SpecialToken(0, '{'),
-                    defs.TextToken(0, 'verbatim'),
-                    defs.SpecialToken(0, '}'),
-        ]
-        self.verbatim_end = parms.scanner.scan('\\end{verbatim}')
-
         # initialise and modify parameters, macros, etc.
-        builtin = [], lambda p: defs.ModParm(
+        builtin = [], lambda p, o: defs.InitModule(
                                 macros_latex=parms.macro_defs_latex,
                                 macros_python=parms.macro_defs_python,
                                 environments=parms.environment_defs)
         for name, actions in [('', builtin)] + packages:
-            self.init_package(name, actions)
+            self.init_package(name, actions, [])
 
     #   call handler of a package module:
     #   - load other required packages
     #   - let package handler update parameter object
     #   - update local dictionaries for macros and environments
+    #   --> only do all that if package not yet loaded, or if options have
+    #       changed
     #
-    def init_package(self, name, actions):
-        if name and name in self.packages:
-            return
+    def init_package(self, name, actions, options):
+        out = []
+        if name and name in self.packages and self.packages[name] == options:
+            return out
         try:
             for requ in actions[0]:
-                if requ not in self.packages:
-                    self.init_package(requ, utils.get_module_handler(
-                                            requ, self.parms.package_modules))
+                if requ not in self.packages or self.packages[requ] == options:
+                    out += self.init_package(requ, utils.get_module_handler(
+                                    requ, self.parms.package_modules), options)
             if name:
-                self.packages.append(name)
-            self.modify_parameters(actions[1])
+                self.packages[name] = options
+            out += self.modify_parameters(actions[1], options)
         except:
             utils.fatal('error loading module ' + repr(name))
+        return out
 
     #   modify local parameter object
     #   - used by package extension mechanism
     #   - used by handlers of LaTeX macros / environments
     #
-    def modify_parameters(self, f):
-        mods = f(self.parms)
+    def modify_parameters(self, f, options):
+        mods = f(self, options)
         for m in mods.macros_python:
             self.the_macros[m.name] = m
         for e in mods.environs:
             self.the_environments[e.name] = e
         if mods.macros_latex:
             self.parser_work(mods.macros_latex)
+        return mods.inject_tokens
 
     #   scan and parse (expand) LaTeX string to tokens
+    #   - here also skip LaTeX text enclosed in special comments
     #
     def parser_work(self, latex):
-        # save self.latex for nested calls, e.g. \LTmacros{}
+        # save self.latex for nested calls, e.g. \LTinput{}
         latex_sav = self.latex
         self.latex = latex
         toks = self.parms.scanner.scan(latex)
+
+        # treat special comments for skipping LaTeX text
+        out = []
+        last = 0
+        while True:
+            beg = next((i for i in range(last, len(toks))
+                    if type(toks[i]) is defs.CommentToken and
+                    toks[i].txt.startswith(self.parms.comment_skip_begin)),
+                                    len(toks))
+            out += toks[last:beg]
+            if beg == len(toks):
+                break
+            end = next((i for i in range(beg + 1, len(toks))
+                    if type(toks[i]) is defs.CommentToken and
+                    toks[i].txt.startswith(self.parms.comment_skip_end)),
+                                    len(toks))
+            if end == len(toks):
+                out += utils.latex_error('cannot find closing LaTeX comment '
+                                    + repr(self.parms.comment_skip_end),
+                                    toks[beg].pos, latex, self.parms)
+                out += toks[beg+1:]
+                break
+            last = end + 1
+        toks = out
+
         toks = self.expand_sequence(scanner.Buffer(toks))
         self.latex = latex_sav
         return toks
 
     #   main entry point
     #
-    def parse(self, latex, extract=None):
+    def parse(self, latex, define='', extract=None):
         if extract:
             self.init_extractions(extract)
         self.extracted = []
         self.unknowns = []
 
-        main = self.parser_work(latex)
+        main = []
+        if define:
+            toks = self.parser_work(define)
+            main = utils.filter_set_toks(toks, 0, defs.LanguageToken)
+        main += self.parser_work(latex)
 
         if extract:
             main = []
@@ -166,7 +192,7 @@ class Parser:
                 t, stop = self.end_environment(buf, tok, env_stop)
                 if stop:
                     return t
-                out += t
+                buf.back(t)
                 continue
             elif type(tok) is defs.ItemToken:
                 buf.back(self.expand_item(buf, tok, out))
@@ -207,14 +233,19 @@ class Parser:
             elif type(tok) is defs.VerbatimToken:
                 if tok.environ:
                     # for Environ() entry in Parameters.environment_defs
-                    t = copy.copy(tok)
-                    t.environ = False
                     buf.next()
-                    buf.back(self.verbatim_begin + [t] + self.verbatim_end)
+                    buf.back(self.expand_verb_env_token(tok))
                     continue
                 else:
                     out.append(defs.ActionToken(tok.pos))
                     out.append(defs.TextToken(tok.pos, tok.txt))
+            elif type(tok) is defs.LanguageToken:
+                if self.parms.multi_language:
+                    self.parms.change_parser_lang(tok)
+                    out.append(tok)
+            elif tok.txt in self.parms.lang_context.active_chars:
+                out.append(self.expand_short_macro(buf, tok))
+                continue
             elif type(tok) is defs.CommentToken:
                 pass
             else:
@@ -315,8 +346,11 @@ class Parser:
             arguments_extr.append(arg_extr)
 
         if mac.extract:
-            toks = self.generate_replacements(arguments_extr,
-                                                        mac.extract, start)
+            toks = ([defs.LanguageToken(start,
+                                    lang=self.parms.lang_context_lang(),
+                                    hard=True, brk=True)]
+                        + self.generate_replacements(arguments_extr,
+                                                        mac.extract, start))
             self.extracted.append(self.expand_sequence(scanner.Buffer(toks)))
         out = [defs.ActionToken(start)]
         if callable(mac.repl):
@@ -414,12 +448,33 @@ class Parser:
                 self.item_lab_stack.pop()
             if env.add_pars:
                 out = [defs.ParagraphToken(tok.pos, '\n\n', pos_fix=True)]
+            if env.end_func:
+                out += env.end_func(self, buf, env, [], tok.pos)
         return out, name == env_stop
 
     def get_environment_name(self, buf, tok):
         buf.next()
         return self.get_text_expanded(self.arg_buffer(buf, tok.pos).all())
 
+    #   expand a single verbatim token delivered by the scanner
+    #   to a valid token sequence for an environment
+    #   --> can be treated like other environments, afterwards
+    #
+    def expand_verb_env_token(self, tok):
+        tok = copy.copy(tok)
+        tok.environ = False
+        return [
+                    defs.BeginToken(tok.pos, '\\begin'),
+                    defs.SpecialToken(tok.pos, '{'),
+                    defs.TextToken(tok.pos, 'verbatim'),
+                    defs.SpecialToken(tok.pos, '}'),
+                    tok,
+                    defs.EndToken(tok.pos + len(tok.txt), '\\end'),
+                    defs.SpecialToken(tok.pos + len(tok.txt), '{'),
+                    defs.TextToken(tok.pos + len(tok.txt), 'verbatim'),
+                    defs.SpecialToken(tok.pos + len(tok.txt), '}'),
+        ]
+
     #   parse (skip) optional [...] after \\
     #
     def parse_newline_option(self, buf, skip_space):
@@ -459,7 +514,8 @@ class Parser:
                 t.can_end = '\n' in txt and not txt[:txt.find('\n')].strip()
             return t
 
-        tokens = [t for t in tokens if t.txt or type(t) is defs.ActionToken]
+        tokens = [t for t in tokens if t.txt or
+                        type(t) in (defs.ActionToken, defs.LanguageToken)]
         tokens = [eval(t) for t in tokens]
         tok = eval(defs.TextToken(0, ''))
         tok.can_start = True
@@ -488,6 +544,7 @@ class Parser:
                     break
             if (can_remove and len(buf) > 1
                     and any(type(t) is defs.ActionToken for t in buf)):
+                lang_toks = [t for t in buf if type(t) is defs.LanguageToken]
                 t1 = copy.copy(buf[0])
                 t2 = copy.copy(buf[-1])
                 # in t1, we remove all behind the last newline
@@ -505,7 +562,7 @@ class Parser:
                 else:
                     t2.txt = ''
                     t2.pos += len(txt)
-                buf = [t1]
+                buf = [t1] + lang_toks
                 tokens.append(eval(t2))
                 # NB: we deleted a line break
                 tok = eval(defs.TextToken(t2.pos, ''))
@@ -515,7 +572,7 @@ class Parser:
                 tokens.append(eval(buf.pop()))
             out += buf
 
-        return [t for t in out if t.txt]
+        return [t for t in out if t.txt or type(t) is defs.LanguageToken]
 
     #   \item: if [...] label is specified, look back in text and append
     #   a possible previous punctuation mark
@@ -542,3 +599,77 @@ class Parser:
         out.insert(0, Space(start))
         return out
 
+    #   expand a short macro like "` for German documents
+    #
+    def expand_short_macro(self, buf, tok):
+        cur = buf.next()
+        if (not cur or tok.txt + cur.txt
+                    not in self.parms.lang_context.short_macros):
+            return tok
+        buf.next()
+        return defs.TextToken(tok.pos,
+                    self.parms.lang_context.short_macros[tok.txt + cur.txt],
+                    pos_fix=True)
+
+    #   parse a key-value list of the form 'a=b,a4paper,x={ hoho }'
+    #   - return a dictionary
+    #   - values are token lists, except if no '=' given: value then is None
+    #   XXX: parsing of key name is not really robust
+    #
+    def parse_keyvals_dict(self, tokens):
+        as_list = self.parse_keyvals_list(tokens)
+        return {k: v for (k, v) in as_list}
+
+    def parse_keyvals_list(self, tokens):
+        buf = scanner.Buffer(tokens)
+        values = []
+        while True:
+            tok = buf.skip_space()
+            if not tok:
+                break
+            key = []
+            while type(tok) is defs.TextToken and tok.txt not in ('=', ','):
+                key.append(tok)
+                tok = buf.next()
+            key = self.get_text_expanded(key)
+            tok = buf.skip_space()
+            if not tok or tok.txt == ',':
+                # no value given with = ...
+                tok = buf.next()
+                values.append((key, None))
+                continue
+            buf.next()                  # skip '='
+            tok = buf.skip_space()      # skip leading space of value
+            val = []
+            while tok and tok.txt != ',':
+                if tok.txt == '{':
+                    # {...} protects space and ','
+                    seq = self.arg_buffer(buf, 0).all()
+                    if len(seq) == 1 and type(seq[0]) is defs.VoidToken:
+                        # this was an empty {}
+                        seq = []
+                    else:
+                        # {} braces have been removed by arg_buffer()
+                        seq = ([defs.SpecialToken(tok.pos, '{')] + seq
+                                    + [defs.SpecialToken(seq[-1].pos, '}')])
+                    val += seq
+                    tok = buf.cur()
+                else:
+                    val.append(tok)
+                    tok = buf.next()
+            if val and type(val[-1]) is defs.SpaceToken:
+                # skip trailing space of value
+                val.pop()
+            values.append((key, val))
+            buf.next()                  # skip ','
+        return values
+
+    #   expand keyval list to text form (None if no '=' given)
+    #
+    def expand_keyvals(self, keyvals):
+        def f(kv):
+            if kv[1] is None:
+                return kv
+            return (kv[0], self.get_text_expanded(kv[1]))
+        return [f(kv) for kv in keyvals]
+
