@@ -23,37 +23,56 @@
 #   differences:
 #   - option '--char' is  always assumed to be set
 #   - option '--defs file' reads macro definitions as LaTeX code
-#   - option '--pack mods' calls functions modify_parameters() from packages
+#   - option '--pack mods' calls functions init_module() from packages
 #
 
 from . import parameters, parser, utils
 
-def tex2txt(latex, opts):
+def tex2txt(latex, opts, multi_language=False, modify_parms=None):
     def read(file):
         try:
             with open(file, encoding=opts.ienc) as f:
                 return True, f.read()
         except:
             return False, ''
-    parms = parameters.Parameters(opts.lang)
+
+    parms = parameters.Parameters(opts.lang or '')
+    parms.multi_language = multi_language
     packages = get_packages(opts.dcls, parms.class_modules)
     packages.extend(get_packages(opts.pack, parms.package_modules))
-    if opts.defs:
-        packages.append(('', utils.get_latex_handler(opts.defs)))
+
     if opts.extr:
         extr = ['\\' + s for s in opts.extr.split(',')]
     else:
         extr = []
+    if opts.seqs:
+        parms.math_displayed_simple = True
+
+    if modify_parms:
+        modify_parms(parms)
     p = parser.Parser(parms, packages, read_macros=read)
-    toks = p.parse(latex, extract=extr)
-    txt, pos = utils.get_txt_pos(toks)
-    if opts.repl:
-        txt, pos = utils.replace_phrases(txt, pos, opts.repl)
-    if opts.unkn:
-        txt = '\n'.join(p.get_unknowns()) + '\n'
-        pos = [0 for n in range(len(txt))]
-    pos = [n + 1 for n in pos]
-    return txt, pos
+    toks = p.parse(latex, define=opts.defs, extract=extr)
+
+    if not multi_language:
+        txt, pos = utils.get_txt_pos(toks)
+        if opts.repl:
+            txt, pos = utils.replace_phrases(txt, pos, opts.repl)
+        if opts.unkn:
+            txt = '\n'.join(p.get_unknowns()) + '\n'
+            pos = [0 for n in range(len(txt))]
+        pos = [n + 1 for n in pos]
+        return txt, pos
+
+    main_lang = opts.lang or ''
+    ml = utils.get_txt_pos_ml(toks, main_lang, parms)
+    if opts.repl and main_lang in ml:
+        for part in ml[main_lang]:
+            part[0], part[1] = utils.replace_phrases(part[0], part[1],
+                                                        opts.repl)
+    for lang in ml:
+        for part in ml[lang]:
+            part[1]= list(n + 1 for n in part[1])
+    return ml
 
 def get_packages(packs, prefix):
     ret = []
@@ -70,6 +89,7 @@ def get_packages(packs, prefix):
             ret.append((p, utils.get_module_handler(p, prefix)))
     return ret
 
+
 #########################################################
 #
 #   the rest is copied from tex2txt/tex2txt.py
@@ -98,6 +118,7 @@ def raise_error(kind, msg, detail=None, xit=None):
     sys.stderr.write(err)
     if xit is not None:
         sys.exit(xit)
+    sys.stderr.flush()
 def strip_internal_marks(s):
     # will be redefined below
     return s
@@ -199,6 +220,7 @@ class Options:
             pack=None,      # import modules for \usepackage
             extr=None,      # or string: comma-separated macro list
             lang=None,      # or set to language code
+            seqs=False,     # True: simple replacements for displayed equations
             unkn=False):    # True: print unknowns
         self.ienc = ienc
         self.repl = repl
@@ -211,6 +233,7 @@ class Options:
         self.pack = pack
         self.extr = extr
         self.lang = lang
+        self.seqs = seqs
         self.unkn = unkn
 
 #   function to be called for stand-alone script
@@ -227,7 +250,9 @@ def main():
     parser.add_argument('--extr')
     parser.add_argument('--lang')
     parser.add_argument('--ienc')
+    parser.add_argument('--seqs', action='store_true')
     parser.add_argument('--unkn', action='store_true')
+    parser.add_argument('--mula')
     cmdline = parser.parse_args()
 
     if not cmdline.ienc:
@@ -242,6 +267,7 @@ def main():
                 pack=cmdline.pack,
                 extr=cmdline.extr,
                 lang=cmdline.lang,
+                seqs=cmdline.seqs,
                 unkn=cmdline.unkn)
 
     if cmdline.file:
@@ -252,6 +278,21 @@ def main():
         # reopen stdin in text mode: handling of '\r', proper decoding
         txt = open(sys.stdin.fileno(), encoding=cmdline.ienc).read()
 
+    if cmdline.mula:
+        # multi-language: write text sections to files
+        ml = tex2txt(txt, options, multi_language=True)
+        for lang in ml:
+            for nr, txt_pos in enumerate(ml[lang]):
+                with myopen(cmdline.mula + '.' + str(nr + 1) + '.' + lang,
+                                    mode='w', encoding='utf-8') as f:
+                    f.write(txt_pos[0])
+                if not cmdline.nums:
+                    continue
+                with myopen(cmdline.nums + '.' + str(nr + 1) + '.' + lang,
+                                    mode='w', encoding='utf-8') as f:
+                    write_output(txt_pos, None, f)
+        sys.exit()
+
     if cmdline.nums:
         cmdline.nums = myopen(cmdline.nums, encoding='utf-8', mode='w')
 
